/* file.c -- Functions to manipulate (copy, move, etc.) files.
   $Id: file.c,v 1.0 2002/04/05 06:51:35 eagle Exp $

   Written by Russ Allbery <rra@stanford.edu>
   Copyright 1997, 1998, 1999, 2000
       Board of Trustees, Leland Stanford Jr. University

   This program is free software.  See README included in the distribution
   for terms and conditions.

   A set of functions to rename, copy, and create files, including doing
   some specialized things we want to do because of the sorts of files that
   we're moving. */

#include "config.h"
#include "file.h"
#include "utility.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <zlib.h>


/* Rename an existing log file so that we can rebind the daemon to a new log
   file.  Logs are always renamed to <log_name>.<time> where <time> is the
   current time in seconds since the epoch.  This should reduce our chances
   of namespace collision to virtually nil.  We return a pointer to the new
   name of the log on success and NULL on failure; memory for the new name
   is malloc()ed and the caller is responsible for freeing it. */
char *
rename_unique(const char *filename)
{
    char *newname;

    /* Allow 20 characters for time to make sure the sun's gone. */
    newname = xmalloc(strlen(filename) + 21);
    sprintf(newname, "%s.%lu", filename, (unsigned long) time(NULL));
    if (rename(filename, newname) != 0) {
        syswarn("Rename of %s failed", filename);
        free(newname);
        return NULL;
    }
    return newname;
}


/* Do the magic of keeping a number of old logs, numbered consecutively
   (such as log.0.gz, log.1.gz, log.2.gz, etc.).  Delete the highest
   numbered log and then rename all logs numbered lower than that to make
   room for a new .0 log. */
int
shuffle_files(const char *filename, int count)
{
    char *logname, *newlogname, *separator, *base;
    int i, remainder;
    int length = 0;

    /* Allocate enough space to store the largest rotation number. */
    assert(count > 0);
    remainder = count;
    for (remainder = count; remainder > 0; remainder /= 10)
        length++;
    logname = xmalloc(strlen(filename) + length);
    base = xmalloc(strlen(filename));

    /* Build the base name (the filename with .0.gz stripped off). */
    strcpy(logname, filename);
    separator = strrchr(logname, '.');
    if (separator) {
        *separator = '\0';
        separator = strrchr(logname, '.');
        if (separator)
            *separator = '\0';
    }
    if (!separator) {
        warn("Invalid archive name %s", filename);
        free(base);
        free(logname);
        return 0;
    }
    strcpy(base, logname);

    /* Remove the highest numbered log, if it exists. */
    sprintf(separator, ".%d.gz", count - 1);
    if (access(logname, F_OK) == 0) {
        if (unlink(logname) != 0) {
            syswarn("Unable to delete %s", logname);
            free(base);
            free(logname);
            return 0;
        }
    }

    /* Rename each log to the next highest number. */
    newlogname = xmalloc(strlen(filename) + length);
    for (i = count - 2; i >= 0; i--) {
        sprintf(logname, "%s.%d.gz", base, i);
        sprintf(newlogname, "%s.%d.gz", base, i + 1);
        if (access(logname, F_OK) == 0) {
            if (rename(logname, newlogname) != 0) {
                syswarn("Unable to rename %s", logname);
                free(base);
                free(logname);
                free(newlogname);
                return 0;
            }
        }
    }

    /* All done.  Clean up the memory we allocated. */
    free(base);
    free(logname);
    free(newlogname);
    return 1;
}


/* Copy a file from one name file to another.  We do this ourselves rather
   than forking a cp process so that we can be sure of the error checking
   (and to avoid external program dependencies). */
int
copy(const char *sourcename, const char *destname)
{
    FILE *source, *dest;
    struct stat statbuf;
    int nread, nwritten, have_size;
    off_t size = 0;
    mode_t mode = 0;

    /* Buffer for transferring the file. */
    static char buffer[2048];

    /* Open the two files. */
    source = fopen(sourcename, "rb");
    if (source == NULL) {
        syswarn("Unable to open source file %s", sourcename);
        return 0;
    }
    dest = fopen(destname, "wb");
    if (dest == NULL) {
        warn("Cannot create %s", destname);
        return 0;
    }

    /* Get the original size of the file in bytes. */
    if (fstat(fileno(source), &statbuf) != 0) {
        syswarn("Unable to stat source file %s", sourcename);
        have_size = 0;
    } else {
        size = statbuf.st_size;
        mode = statbuf.st_mode;
        have_size = 1;
    }

    /* If we can, chmod the archived log to the same as the original. */
    if (have_size && fchmod(fileno(dest), mode) != 0)
        syswarn("Unable to chmod destination file %s", destname);

    /* Copy the file over. */
    while (!feof(source)) {
        nread = fread(buffer, 1, 2047, source);
        nwritten = fwrite(buffer, 1, nread, dest);
        if (nread != nwritten || ferror(source)) {
            warn("Error while writing %s", destname);
            fclose(source);
            fclose(dest);
            return 0;
        }
    }

    /* Close the files and then verify that the copy was successful. */
    fclose(source);
    if (fclose(dest) != 0) {
        warn("Error while writing %s", destname);
        return 0;
    }
    if (!have_size || stat(destname, &statbuf) != 0)
        warn("Unable to verify successful copy of %s", sourcename);
    else if (size != statbuf.st_size) {
        warn("Error copying %s", sourcename);
        warn("  size of copy (%lu) != size of original (%lu)",
             statbuf.st_size, size);
        return 0;
    }

    return 1;
}


/***************************************************************************
* Creating files and directories
***************************************************************************/

/* Create a new file with the given name, owner, group, and permissions. */
int
create_file(const char *file, uid_t owner, gid_t group, mode_t mode)
{
    int fd;

    fd = open(file, O_CREAT | O_WRONLY, mode);
    if (fd < 0) {
        syswarn("Creation of %s failed", file);
        return 0;
    }
    if (fchown(fd, owner, group) != 0)
        syswarn("Unable to chown %s", file);
    if (fchmod(fd, mode) != 0)
        syswarn("Unable to chmod %s", file);
    close(fd);
    return 1;
}


/* Creates the parent directory of a file, if it doesn't already exist. */
void
create_directory(char *path)
{
    char *separator;

    separator = strrchr(path, '/');
    if (separator) {
        *separator = '\0';
        if (access(path, W_OK) != 0)
            if (mkdir(path, 0755) != 0 && errno != EEXIST)
                syswarn("Could not create directory %s", path);
        *separator = '/';
    }
}


/***************************************************************************
* Compressing logs
***************************************************************************/

/* Copy a file from one location to another, compressing it in the process.
   The preferred method is to use zlib.  If we can't do that, we'll have to
   use a pipe to a gzip process (not currently implemented). */
int
compress_file(const char *sourcename, const char *destname)
{
    FILE *source;
    gzFile *dest;
    struct stat statbuf;
    int fd, nread, nwritten, have_mode;

    /* Buffer for transferring the file. */
    static char buffer[2048];

    /* Open the source file as a standard file and get its mode. */
    source = fopen(sourcename, "rb");
    if (!source) {
        syswarn("Unable to open source file %s", sourcename);
        return 0;
    }
    if (fstat(fileno(source), &statbuf) == 0) {
        have_mode = 1;
    } else {
        syswarn("Unable to stat source file %s", sourcename);
        have_mode = 0;
    }

    /* Open the destination file and try to chmod it to the right perms. */
    fd = open(destname, O_CREAT | O_EXCL | O_WRONLY, 0600);
    if (fd < 0) {
        syswarn("Cannot create %s", destname);
        fclose(source);
        return 0;
    }
    if (have_mode && fchmod(fd, statbuf.st_mode) != 0)
        syswarn("Unable to chmod destination file %s", destname);

    /* Now, open this file descriptor as a maximally compressed file. */
    dest = gzdopen(fd, "wb9");
    if (dest == NULL) {
        warn("Cannot begin compressing to %s", destname);
        fclose(source);
        close(fd);
        return 0;
    }

    /* Copy the file over.  Maybe some day add better error reporting. */
    while (!feof(source) && !ferror(source)) {
        nread = fread(buffer, 1, 2047, source);
        nwritten = gzwrite(dest, buffer, nread);
        if (nread != nwritten || ferror(source)) {
            warn("Error while writing %s", destname);
            fclose(source);
            gzclose(dest);
            return 0;
        }
    }

    /* Close the files and then verify that the copy was successful. */
    fclose(source);
    if (gzclose(dest) != 0) {
        warn("Error while writing %s", destname);
        return 0;
    }

    return 1;
}
